'use strict';

/**
 * Check if RTSP Viewer user exists on camera and a valid pass is set in Camstreamer.CSUSerPass parameter
 *
 */

/**
 * @return {Readonly<{checkViewer: checkViewer}>}
 * @constructor
 */
let UserManager = function() {
  async function checkViewer() {
    let current_pass = await getCurrentPass();
    let pass_ok = await isPassValid(current_pass);
    if (pass_ok) {
      return true;
    }

    let username = 'camstreamer';
    let pass = randomPassword(16);
    if (await userExists(username)) {
      await updateUserPass(username, pass);
      await updatePassParameter(pass);
    } else {
      await createUser(username, pass);
      await updatePassParameter(pass);
    }
  }

  function randomPassword(length) {
    const validChars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
    let array = new Uint8Array(length);
    window.crypto.getRandomValues(array);
    array = array.map(x => validChars.charCodeAt(x % validChars.length));
    const password = String.fromCharCode.apply(null, array);
    return password;
  }

  async function isPassValid(pass) {
    let response = await $.getJSON('/local/camstreamer/check_pass.cgi', {pass});
    return response.data === '1';
  }

  async function userExists(username) {
    let response = await $.get('/axis-cgi/pwdgrp.cgi?action=get');
    let users_string = response.match(/^users="([a-z0-9,]*)"/mi)[1];
    let users = users_string.split(',');
    let viewers_string = response.match(/^viewer="([a-z0-9,]*)"/mi)[1];
    let viewers = viewers_string.split(',');

    if (users.includes(username) && viewers.includes(username)) {
      return true;
    }

    if (users.includes(username) && !viewers.includes(username)) {
      await removeUser(username);
    }
    return false;
  }

  async function createUser(username, pass) {
    let params = {
      user: username,
      pwd: pass,
      grp: 'users',
      sgrp: 'viewer',
      comment: 'Camstreamer RTSP viewer',
    };
    let response = await $.get('/axis-cgi/pwdgrp.cgi?action=add', params);
    let error_match = response.match(/Error:([^<]*)/);
    if (error_match !== null) {
      throw Error("Failed to create user: " + error_match[1].trim());
    }
  }

  async function removeUser(username) {
    let response = await $.get('/axis-cgi/pwdgrp.cgi?action=remove', {user: username});
    let error_match = response.match(/Error:([^<]*)/);
    if (error_match !== null) {
      throw Error("Failed to remove user: " + error_match[1].trim());
    }
  }

  async function updateUserPass(username, pass) {

    let update_user_response = await $.get('/axis-cgi/pwdgrp.cgi?action=update', {user: username, pwd: pass});
    let error_match = update_user_response.match(/Error:([^<]*)/);
    if (error_match !== null) {
      throw Error("Failed to update user pass: " + error_match[1].trim());
    }
  }

  async function updatePassParameter(pass) {
    let update_params_response = await $.post('/axis-cgi/param.cgi', {action: 'update', 'Camstreamer.CSUSerPass': pass});
    if (update_params_response !== 'OK') {
      throw Error("Failed to update CSUSerPass parameter: " + update_params_response.trim());
    }
  }

  async function getCurrentPass() {
    let response = await $.get('/axis-cgi/param.cgi?action=list', {group: 'Camstreamer.CSUSerPass'});
    let pass_match = response.match(/^.*?=(.*)$/m);
    if (pass_match === null) {
      throw Error('Bad response');
    }
    return pass_match[1];
  }

  return Object.freeze({
    checkViewer,
    randomPassword,
  });
}();

