/**
 * Handles postMessage API on camera from ACAP from version CS 3.3
 */

'use strict';

const CS_OS = {
  acap_name: 'camstreamer',
  allowed_eventmessage_origin: null,
  ui_domain: null,
  cameraWebsockets: websocketClient(window.location.host, window.location.protocol === 'https:'),
  firmware_version: null,
};

const alertBox = AlertBox($('#alertbox'));
const serverIframe = $("#server_iframe");
const camera = Camera();

window.addEventListener('message', receiveMessage, false);

// replace "Loading UI" message with proper one with category. We want to use 'hide' instead of 'clear' later
alertBox.clear();
alertBox.showInfo('Loading UI...', 'loading');

checkAcapRunning().catch( function (e) {
  serverIframe.remove(); // UI is loading simultaneously - hide it.
  alertBox.showError(`Can’t verify CamStreamer ACAP is running - ${e.message}. Please refresh this page or check manually in Applications.`);
});

init().catch(function (e) {
  alertBox.showError(e.message);
});

async function init()
{
  const params = await fetchParameters();
  const serverUrl = params.serverUrl;

  CS_OS.allowed_eventmessage_origin = serverUrl;
  CS_OS.ui_domain = serverUrl;
  CS_OS.firmware_version = params.camera_firmware;

  const iframeUrl = buildIframeUrl(serverUrl, params);
  alertBox.hide('loading');
  serverIframe.attr('src', iframeUrl);
  serverIframe.show();

  renderMenu('', serverUrl);

  $('.ui_domain_link').each( (index, element) => element.href = element.href.replace('https://camstreamer.com', serverUrl));

  UserManager.checkViewer().catch( err => console.error('Failed to check CamStreamer RTSP Viewer', err));
}

async function fetchParameters()
{
  const param_names = {
    serverUrl: 'Camstreamer.ServerURL',
    serial_number: 'Properties.System.SerialNumber',
    camera_model: 'Brand.ProdNbr',
    camera_firmware: 'Properties.Firmware.Version',
    acap_version: 'Camstreamer.Version',
    licensecode: 'Camstreamer.Licensecode',
    timezone: 'Time.POSIXTimeZone',
  };
  const params = await camera.getParameters(param_names);

  const invalidServerUrlMessageFactory = function (reason) {
    return `An invalid <i>Server URL</i> is set. ${reason} Please change the URL to https://camstreamer.com by following our <a href="https://camstreamer.com/blog/how-to-set-up-correct-server-url-to-camstreamer-application" target="_blank">setup guide</a>.`;
  }

  let serverUrl = String(params.serverUrl);
  if (serverUrl.charAt(serverUrl.length-1) === '/') {
    throw new Error(invalidServerUrlMessageFactory("URL should not end with /."));
  }

  let url;
  try {
    url = new URL(serverUrl);
  } catch (e) {
    throw new Error(invalidServerUrlMessageFactory("Value is not an URL."));
  }

  if (url.protocol !== 'https:') {
    throw new Error(invalidServerUrlMessageFactory("Value must begin with 'https:' - SSL is required."));
  }

  return params;
}

function buildIframeUrl(domain, params) {
  const httpport = location.port ? ':' + location.port : '';
  params.baseurl = location.protocol + '//' + location.hostname + httpport;

  if (window.location.hash.match(/^#(\d+)$/)) {
    params.a = 'editStream';
    params.stream_id = window.location.hash.match(/^#(\d+)$/)[1];
  } else {
    params.a = 'default';
  }
  let url = domain + "/index.php?p=ACAPLanding";
  for (let param_name in params) {
    url += '&' + param_name + '=' + encodeURIComponent(params[param_name]);
  }

  return url;
}

/**
 * Make sure CamStreamer ACAP is running.
 */
async function checkAcapRunning() {
  return $.get("/axis-cgi/applications/control.cgi?action=start&package=" + CS_OS.acap_name)
   .catch(function (jqXHR, textStatus, errorThrown) {
     console.error(textStatus, errorThrown);
     throw new Error('Network error: ' + errorThrown || textStatus);
   })
   .then(function (response) {
     if (typeof response !== 'string') {
       throw new Error('No result');
     }
     if (response.indexOf('OK') !== -1) {
       // ACAP started successfully
     } else if (response !== 'Error: 6') { //ACAP is already running
       throw new Error(response);
     }
   });
}

async function receiveMessage(event) {
  if (event.origin !== CS_OS.allowed_eventmessage_origin) {
    return;
  }

  let event_data;
  // forward compatibility - we might remove explicit data json encoding //@TODO remove after release-2.4
  if (typeof event.data === 'string') {
    try {
      event_data = JSON.parse(event.data);
    } catch (e) {
      return;
    }
  } else {
    event_data = event.data;
  }

  if (typeof event_data !== 'object') {
    return;
  }

  const reply = function (error, response) {
    if (error instanceof Error) {
      error = `${error.name}: ${error.message}`;
    }
    const reply_data = {
      queryid: event_data.queryid,
      error: error,
      response: response,
    };
    event.source.postMessage(reply_data, CS_OS.allowed_eventmessage_origin);
  };

  const sendError = function (error) {
    reply(error, null);
  };

  const sendData = function (data) {
    reply(null, data);
  };
  
  if (event_data.query === "getCameraParameters") {
    camera.getParameters(event_data.parameters)
      .then(sendData)
      .catch( error => sendError("Failed to get parameters: " + error.message));
  }

  if (event_data.query === "setCameraParameters") {
    camera.setParameters(event_data.parameters)
      .then(x => sendData(true))
      .catch( error => sendError("Failed to set parameters: " + error.message));
  }

  if (event_data.query === "getConnectionDiagnostics") {
    event_data.params['dummy'] = new Date().getTime();
    $.get("/local/" + CS_OS.acap_name + "/diagnostics.cgi?" + buildQueryString(event_data.params), function (response) {
      sendData(response.data);
    }).fail(function( jqXHR, textStatus, errorThrown ){
      sendError(errorThrown || 'Network error');
    });
  }

  if (event_data.query === "getStreamStat") {
    var stream_id = event_data.hasOwnProperty('stream_id') ? event_data.stream_id : "";
    let params = {
      stream_id: stream_id,
      dummy: new Date().getTime(),
    };
    $.getJSON("/local/" + CS_OS.acap_name + "/get_streamstat.cgi", params)
      .done(function (response, status, jqXHR) {
        if (response.status !== 200) {
          sendError(response.message);
          return;
        }
        let stats = response.data;

        if (typeof stats === 'object' && stats.hasOwnProperty('net_stats') && typeof stats['net_stats'] === 'string') {
          sendData(stats);
        } else {
          sendError('Missing net_stats data');
        }
      })
      .fail(function(jqXHR, textStatus, errorThrown) {
        sendError(errorThrown || 'Network error');
      });
  }

  if (event_data.query === "getSrtStats") {
    if (!event_data.hasOwnProperty('stream_id')) {
      sendError('Missing stream_id parameter.');
    }
    const stream_id = event_data.stream_id;
    $.getJSON("/local/" + CS_OS.acap_name + "/srt_statistics.cgi", {stream_id: stream_id}).then(function (response) {
      if (response.status === 404 || !!response.data && Object.values(response.data).length === 0) {
        sendError('NOT_READY');
      } else if (response.status !== 200) {
        sendError(response.message);
      } else {
        sendData(response.data);
      }

    }).fail(function(jqXHR, textStatus, errorThrown) {
      sendError(errorThrown || 'Network error');
    });
  }

  if (event_data.query === "rebootCamera") {
    $.get("/axis-cgi/restart.cgi?dummy="+(new Date().getTime()));
  }

  if (event_data.query === "restartAcap") {
    $.get(`/axis-cgi/applications/control.cgi?action=restart&package=${CS_OS.acap_name}`)
      .then(function (response) {
        if (response === 'Error: 7') {
          $.get(`/axis-cgi/applications/control.cgi?action=start&package=${CS_OS.acap_name}`);
        }
      })
      .catch(function () {
        console.error('Failed to restart Camstreamer App.');
      });
  }

  if (event_data.query === "getMaxAvailableFps") {
    let video_source = parseInt(event_data.video_source, 10) || 0;

    if (CS_OS.firmware_version === null || firmwareVersionCompare(CS_OS.firmware_version, '8.45.3') === -1) {
      CamstreamerAPI().getMaxFps(video_source).then(sendData, sendError);
    } else {
      // capturemode.cgi is officially supported from FW 8.50 but is must be used on V59x from 8.45.3 too
      camera.getMaxFps(video_source).then(sendData, sendError);
    }
  }

  if (event_data.query === "getUtcTime") {
    $.get("/local/" + CS_OS.acap_name + "/get_utc_time.cgi?dummy="+(new Date().getTime()))
      .done( function (response) {
        let timestamp = parseInt(response);
        if (Number.isInteger(timestamp)) {
          sendData(timestamp);
        } else {
          sendError('INVALID_TIMESTAMP');
        }
      })
      .fail( function(jqXHR, textStatus, errorThrown) {
        sendError(errorThrown || 'Network error');
      });
  }

  if (event_data.query === "getAudioState") {
    $.getJSON("/local/" + CS_OS.acap_name + "/get_audiostate.cgi?dummy="+(new Date().getTime()))
      .done( function (response) {
        sendData(response);
      })
      .fail( function(jqXHR, textStatus, errorThrown) {
        sendError(errorThrown || 'Network error');
      });
  }

  if (event_data.query === "listAudioFiles") {
    $.getJSON("/local/" + CS_OS.acap_name + "/upload_audio.cgi?action=list&dummy="+(new Date().getTime()))
      .done( function (response) {
        let fileList = [];
        $.each(response, function(id, file) {
          let basename = file.replace(/.*\//, "" );
          fileList.push({ path: file, basename: basename});
        });
        sendData(fileList);
      })
      .fail( function(jqXHR, textStatus, errorThrown) {
        sendError(errorThrown || 'Network error');
      });
  }

  if (event_data.query === "reload") {
    location.reload();
  }

  if (event_data.query === "getStream") {
    if (event_data.hasOwnProperty('stream_id') === false || !event_data.stream_id) {
      sendError('stream_id must be set');
      return;
    }
    if (typeof event_data.params !== 'object') {
      event_data.params = { };
    }
    event_data.params['stream_id'] = event_data.stream_id;
    event_data.params['dummy'] = new Date().getTime();

    let response;
    try {
      response = await fetch("/local/" + CS_OS.acap_name + "/stream/get.cgi?" + buildQueryString(event_data.params));
      if (!response.ok) {
        throw Error('Communication error');
      }
      response = await response.json();
    } catch (error) {
      sendError(error || 'Network error');
    }

    if (response.status === 200) {
      sendData(response.data);
    } else if (response.status === 404) {
      sendError(`Stream not found`);
    } else {
      sendError(response.message || String(response.status));
    }
  }

  if (event_data.query === "setStream") {
    if (event_data.hasOwnProperty('stream_id') === false || !event_data.stream_id) {
      sendError('stream_id must be set');
      return;
    }
    if (typeof event_data.params !== 'object') {
      event_data.params = { };
    }
    event_data.params['stream_id'] = event_data.stream_id;
    event_data.params['dummy'] = new Date().getTime();
    let request = new Request("/local/" + CS_OS.acap_name + "/stream/set.cgi", {
      method: 'POST',
      headers: {
        'Content-Type': 'application/x-www-form-urlencoded',
      },
      body: buildQueryString(event_data.params),
    });

    let response;
    try {
      response = await fetch(request);
      if (!response.ok) {
        throw Error('Communication error');
      }
      response = await response.json();
    } catch (error) {
      sendError(error || 'Network error');
    }

    if (response.status === 200) {
      sendData(response.data);
    } else if (response.status === 404) {
      sendError(`Stream not found`);
    } else {
      sendError(response.message || String(response.status));
    }
  }

  if (event_data.query === "removeStream") {
    if (event_data.hasOwnProperty('stream_id') === false || !event_data.stream_id) {
      sendError('stream_id must be set');
      return;
    }
    if (typeof event_data.params !== 'object') {
      event_data.params = { };
    }
    event_data.params['stream_id'] = event_data.stream_id;
    event_data.params['dummy'] = new Date().getTime();
    $.getJSON("/local/" + CS_OS.acap_name + "/stream/remove.cgi?" + buildQueryString(event_data.params))
      .done( function (response) {
        if (response.status === 200) {
          sendData(response.data);
        } else {
          sendError(response.message || String(response.status));
        }
      }).fail( function(jqXHR, textStatus, errorThrown) {
        sendError(errorThrown || 'Network error');
      });
  }

  if (event_data.query === "listStreams") {
    let params = {
      dummy: new Date().getTime(),
    };
    $.getJSON("/local/" + CS_OS.acap_name + "/stream/list.cgi?" + buildQueryString(params))
      .done( function (response) {
        if (response.status === 200) {
          sendData(response.data);
        } else {
          sendError(response.message || String(response.status));
        }
      }).fail( function(jqXHR, textStatus, errorThrown) {
      sendError(errorThrown || 'Network error');
    });
  }

  if (event_data.query === "navigateStream") {
    window.location.hash = parseInt(event_data.stream_id) || '';
  }

  if (event_data.query === "getServerReport") {
    try {
      const mode = event_data.mode || 'text';
      const report = await camera.getServerReport(mode);
      sendData(report);
    } catch (err) {
      sendError(err.message);
    }
  }

  if (event_data.query === "getDiagnosticsReport") {
    let attachments = event_data.attachments;
    try {
      let report = await camera.getReport(attachments);
      sendData(report);
    } catch (err) {
      sendError(err.message);
    }
  }

  if (event_data.query === "updateAcap") {
    CS_OS.cameraWebsockets.close();
    try {
      await getUpdate('camstreamer','CamStreamer', CS_OS.ui_domain, installPage, defaultPage);
    } catch (error) {
      sendError(error.message);
    }
    sendData('OK');
  }

  if (event.data.query === 'reconnectCameraWebsocket') {
    CS_OS.cameraWebsockets.sendInitData();
  }

  if (event.data.query === 'closeCameraWebsocket') {
    CS_OS.cameraWebsockets.close();
  }

  if (event.data.query === 'registerCameraWebsocketListener') {
    CS_OS.cameraWebsockets.addPushMessageTarget(event.source, event.origin);
  }

  if (event.data.query === 'record/rule/list') {
    try {
      const list = await RecordingAPI().ruleList();
      sendData(list);
    } catch (e) {
      sendError(e.message);
    }
  }

  if (event.data.query === 'record/rule/add') {
    const diskid = event.data.diskid;
    const options = event.data.options;
    try {
      const profile_id = await RecordingAPI().ruleAdd(diskid, options);
      sendData(profile_id);
    } catch (e) {
      sendError(e.message);
    }
  }

  if (event.data.query === 'record/rule/remove') {
    const profile_id = event.data.profile_id;
    if (!profile_id && profile_id !== 0) {
      sendError('No profile id given.');
    }
    try {
      await RecordingAPI().ruleRemove(profile_id);
      sendData(true);
    } catch (e) {
      sendError(e.message);
    }
  }

  if (event.data.query === 'edgestorage/isReady') {
    const diskid = event.data.diskid;
    if (!diskid) {
      sendError('No disk id given.');
    }
    try {
      const is_available = await RecordingAPI().checkEdgeReady(diskid);
      sendData(is_available);
    } catch (e) {
      sendError(e.message);
    }
  }

}

/**
 * Convert object parameters to query string. Correctly encoding spaces to %20 instead of '+'
 * @param {object} params query parameters
 * @returns {string}
 */
function buildQueryString(params) {
  if (typeof params !== 'object') {
    return '';
  }
  var query = [];
  for (var name in params) {
    query.push(name + "=" + encodeURIComponent(params[name]));
  }
  return query.join('&');
}

/**
 * @param {string} a
 * @param {string} b
 * @return {number} [-1, 0, 1] @see spaceship/combined comparsion operator
 */
function firmwareVersionCompare(a, b)
{
  const versions = [a, b];
  if (versions.some(x => typeof x !== 'string')) {
    throw new Error(`At least one of versions is not of a string type.`);
  }

  const matchBetaFirmwareVersion = x => /^CVP-/.test(x) || /^[0-9]+.*beta/.test(x);
  if (versions.some(matchBetaFirmwareVersion)) {
    if (versions.every(matchBetaFirmwareVersion)) {
      return 0; // both versions are beta
    }
    if (matchBetaFirmwareVersion(a)) {
      return -1;
    }
    if (matchBetaFirmwareVersion(b)) {
      return 1;
    }
  }
  return versionCompare(a, b);
}

/**
 *
 * @param {string} a
 * @param {string} b
 * @return {number} [-1, 0, 1] @see spaceship/combined comparsion operator
 */
function versionCompare(a, b)
{
  [a, b].forEach(function (x) {
    if (typeof x !== "string" || !x.match(/^[0-9]+(\.[0-9]+){1,3}$/)) {
      throw new Error(`Version string '${x}' is not in allowed format.`);
    }
  });

  if (a === b) {
    return 0;
  }
  const a_split = a.split('.', 4);
  const b_split = b.split('.', 4);
  for (let i = 0; i < 4; i++) {
    let a_curr = a_split[i] || 0;
    let b_curr = b_split[i] || 0;
    if (a_curr !== b_curr) {
      return a_curr < b_curr ? -1 : 1;
    }
  }
  throw new Error(`Failed to compare '${a}' with '${b}'`); // should not happen...
}